package structr

import "unsafe"

type structA struct {
	Field1 string
	Field2 int
	Field3 float32
}

var structAIndices = []IndexConfig{
	{
		Fields:    "Field1",
		Multiple:  false,
		AllowZero: false,
	},
	{
		Fields:    "Field2",
		Multiple:  true,
		AllowZero: true,
	},
	{
		Fields:    "Field3",
		Multiple:  true,
		AllowZero: true,
	},
	{
		Fields:    "Field2,Field3",
		Multiple:  false,
		AllowZero: true,
	},
}

var structAValues = []*structA{
	{Field1: "zero-zero", Field2: 0, Field3: 0},
	{Field1: "one-zero", Field2: 1, Field3: 0},
	{Field1: "zero-one", Field2: 0, Field3: 1},
	{Field1: "69-420", Field2: 69, Field3: 420},
	{Field1: "420-69", Field2: 420, Field3: 69},
}

var testStructA = test[structA]{
	indices: structAIndices,
	copyfn: func(in *structA) *structA {
		out := new(structA)
		*out = *in
		return out
	},
	values: structAValues,
	equalfn: func(a, b *structA) bool {
		return a.Field1 == b.Field1 &&
			a.Field2 == b.Field2 &&
			a.Field3 == b.Field3
	},
}

type structB struct {
	Field1 *string
	Field2 *int
	Field3 *float64
}

var structBIndices = []IndexConfig{
	{
		Fields:    "Field1",
		Multiple:  false,
		AllowZero: false,
	},
	{
		Fields:    "Field2",
		Multiple:  true,
		AllowZero: true,
	},
	{
		Fields:    "Field3",
		Multiple:  true,
		AllowZero: true,
	},
	{
		Fields:    "Field2,Field3",
		Multiple:  false,
		AllowZero: true,
	},
}

var structBValues = []*structB{
	{Field1: ptr("nil-nil"), Field2: nil, Field3: nil},
	{Field1: ptr("one-nil"), Field2: ptr(1), Field3: nil},
	{Field1: ptr("nil-one"), Field2: nil, Field3: ptr(1.0)},
	{Field1: ptr("69-420"), Field2: ptr(69), Field3: ptr(420.0)},
	{Field1: ptr("420-69"), Field2: ptr(420), Field3: ptr(69.0)},
}

var testStructB = test[structB]{
	indices: structBIndices,
	copyfn: func(in *structB) *structB {
		out := new(structB)
		*out = *in
		return out
	},
	values: structBValues,
	equalfn: func(a, b *structB) bool {
		return ptrsequal(a.Field1, b.Field1) &&
			ptrsequal(a.Field2, b.Field2) &&
			ptrsequal(a.Field3, b.Field3)
	},
}

type structC struct {
	Field1 string
	Field2 *int
	Field3 *struct {
		Field1 string
		Field2 *int
	}
}

var structCIndices = []IndexConfig{
	{
		Fields:    "Field1",
		Multiple:  false,
		AllowZero: false,
	},
	{
		Fields:    "Field2",
		Multiple:  true,
		AllowZero: true,
	},
	{
		Fields:    "Field3.Field1",
		Multiple:  true,
		AllowZero: true,
	},
	{
		Fields:    "Field3.Field2",
		Multiple:  true,
		AllowZero: true,
	},
}

var structCValues = []*structC{
	{Field1: "zero-zero", Field2: nil, Field3: &struct {
		Field1 string
		Field2 *int
	}{
		Field1: "zero-zero",
		Field2: nil,
	}},
	{Field1: "one-zero", Field2: ptr(1), Field3: &struct {
		Field1 string
		Field2 *int
	}{
		Field1: "one-zero",
		Field2: ptr(1),
	}},
	{Field1: "zero-one", Field2: nil, Field3: &struct {
		Field1 string
		Field2 *int
	}{
		Field1: "zero-one",
		Field2: nil,
	}},
	{Field1: "69-420", Field2: ptr(69), Field3: &struct {
		Field1 string
		Field2 *int
	}{
		Field1: "69-420",
		Field2: ptr(69),
	}},
	{Field1: "420-69", Field2: ptr(420), Field3: &struct {
		Field1 string
		Field2 *int
	}{
		Field1: "420-69",
		Field2: ptr(420),
	}},
	{Field1: "empty-nil", Field2: nil, Field3: nil},
}

var testStructC = test[structC]{
	indices: structCIndices,
	copyfn: func(in *structC) *structC {
		out := new(structC)
		*out = *in
		if in.Field3 != nil {
			out.Field3 = new(struct {
				Field1 string
				Field2 *int
			})
			*out.Field3 = *in.Field3
		}
		return out
	},
	values: structCValues,
	equalfn: func(a, b *structC) bool {
		if a.Field1 != b.Field1 ||
			!ptrsequal(a.Field2, b.Field2) {
			return false
		}
		if a.Field3 == nil {
			return b.Field3 == nil
		}
		if b.Field3 == nil {
			return false
		}
		return a.Field3.Field1 == b.Field3.Field1 &&
			ptrsequal(a.Field3.Field2, b.Field3.Field2)
	},
}

// indexkey extracts the index key interface parts for index.
func indexkey[T any](index *Index, value *T) ([]any, bool) {
	parts := extract_fields(unsafe.Pointer(&value), index.fields)
	return parts, (parts != nil)
}

func ptr[T any](t T) *T { return &t }

func ptrsequal[T comparable](p1, p2 *T) bool {
	switch {
	case p1 == nil:
		return (p2 == nil)
	case p2 == nil:
		return false
	default:
		return (*p1 == *p2)
	}
}

func catchpanic(do func(), expect any) {
	defer func() {
		r := recover()
		if r != expect {
			panicf("expected panic with %v, got %v", expect, r)
		}
	}()
	do()
}
